import inspect
from typing import (
    Awaitable,
    Callable,
    Protocol,
    Union,
)


from .structs import (
    Generate,
    User,
)

from .turbo import turbo

from .utils.args import ensure_args


__all__ = [
    "completion",
]


class CompletionFnWithResult(Protocol):
    def __call__(self, result: str, **kwargs) -> Awaitable[str]:
        ...


class CompletionFnWithoutResult(Protocol):
    def __call__(self, **kwargs) -> Awaitable[None]:
        ...


CompletionFn = Union[CompletionFnWithResult, CompletionFnWithoutResult]


class Completion(Protocol):
    def __call__(self, **kwargs) -> Awaitable[str]:
        ...


# Decorator
def completion(**opts) -> Callable[[CompletionFn], Completion]:
    """Parameterized decorator for creating a simple generate function"""

    def wrapper(fn: CompletionFn):
        @turbo(**opts)
        async def generate(**kwargs):
            yield User(template=inspect.getdoc(fn), variables=kwargs)
            yield Generate()

        # @wraps(fn)
        async def wrapped(**kwargs) -> str:
            # Ensure args
            assert ensure_args(fn, kwargs)

            # Get result
            result = await generate(**kwargs).run()

            return result.content

        return wrapped

    return wrapper



from typing import cast, Generic, List, TypeVar

import dirtyjson
from parse import compile as compile_parser, Parser, with_pattern

__all__ = [
    "Scratchpad",
]

ST = TypeVar("ST")  # , bound=TypedDict) # Not available in py3.8

yesno_mapping = {
    "yes": True,
    "no": False,
    "on": True,
    "off": False,
    "true": True,
    "false": False,
    "1": True,
    "0": False,
}


@with_pattern(r"|".join(yesno_mapping))
def parse_yesno(text: str) -> bool:
    return yesno_mapping[text.lower()]


@with_pattern(r"(\{.+\})")
def parse_json(text: str) -> dict:
    return dirtyjson.loads(text)


@with_pattern(r".+")
def parse_multiline(text: str) -> str:
    return text


class Scratchpad(Generic[ST]):
    parsers: List[Parser]

    def __init__(self, spec: str):
        extra_types = dict(bool=parse_yesno, json=parse_json, multiline=parse_multiline)

        self.parsers = [
            compile_parser(
                spec_line + ("" if ":multiline" in spec_line else "\n"),
                extra_types=extra_types,
            )
            for spec_line in spec.strip().splitlines()
        ]

    def parse(self, input: str) -> ST:
        """Parse the input string according to the spec."""

        # Parse the scratchpad
        result = {}
        for parser in self.parsers:
            parsed = parser.search(input + "\n")
            result = {**result, **getattr(parsed, "named", {})}

        # Trim results
        result = {k: v.strip() if isinstance(v, str) else v for k, v in result.items()}

        return cast(ST, result)



from typing import Awaitable, Callable, Optional, TypedDict

__all__ = [
    "Tool",
    "ToolBotInput",
]

Tool = Callable[..., Awaitable[str]]


class ToolBotInput(TypedDict):
    state: Optional[str]
    input: str



from typing import List, Literal, Optional

from ...structs import Example, GetInput, Generate, Result, User
from ...turbo import turbo
from ...types import Tool, ToolBotInput
from .scratchpad import scratchpad
from .template import TOOLBOT_TEMPLATE

default_tool_example: str = """
Customer said: Where is the store located?
Thought: Use another tool to get information? Yes
Tool: GetInformation
Tool Input: {"question": "What is the address of the store?"}
Tool Result: The address of [STORE NAME] is [SOME ADDRESS]
Thought: Use another tool to get information? No
Response: [STORE NAME] is located at [SOME ADDRESS].
""".strip()


@turbo(temperature=0)
async def tool_bot(
    tools: List[Tool],
    prologue: Optional[str] = None,
    user_type: str = "user",
    instruction: Optional[str] = None,
    initial_state: Optional[str] = None,
    example: str = default_tool_example,
    max_iterations: int = 6,
    state_message_position: Literal["top", "bottom"] = "top",
    # Placeholder for message to send to the user
    message: str = "How can I help you?",
):
    # Capitalize user_type
    user_type = user_type.strip().capitalize()

    # Yield initial state
    if initial_state:
        yield User(
            content=initial_state,
            sticky=True,
            sticky_position=state_message_position,
            label="state",
        )

    # Yield instructions
    yield User(
        template=TOOLBOT_TEMPLATE,
        variables=dict(
            tools=tools,
            prologue=prologue,
            user_type=user_type,
            instruction=instruction,
        ),
        # check=True,  # FIXME: Throws an error, possible bug in jinja2schema
    )

    # Yield example
    # yield User(template=EXAMPLE_TEMPLATE, variables={"example": example})
    example_user_line, *example_assistant_lines = example.splitlines()
    yield Example(
        user=example_user_line,
        assistant="\n".join(example_assistant_lines),
    )

    while True:
        # Get user input
        data: ToolBotInput = yield GetInput(content=message)
        input: str = data["input"]
        state: Optional[str] = data.get("state")

        # Yield user input and state
        if state:
            yield User(
                content=state,
                sticky=True,
                sticky_position=state_message_position,
                label="state",
            )

        yield User(content=f"{user_type} said: {input}")

        # Start the tool agent
        parsed_tools = {}
        iterations_left = max_iterations
        tools_used = []

        # Keep running until we get final response
        while "final_response" not in parsed_tools and iterations_left > 0:
            iterations_left -= 1

            output = yield Generate(
                stop=["Tool Result", f"{user_type} said:"],
                forward=False,
            )

            parsed_tools = scratchpad.parse(output.content)

            # If no tool required to run, continue
            if not parsed_tools.get("tool_name"):
                continue

            # Otherwise run tool
            # First check if tool is valid
            selected_tool = parsed_tools["tool_name"]
            tool_names = [tool.__name__ for tool in tools]

            if selected_tool not in tool_names:
                yield User(
                    content=(
                        f"Tool Result: `{selected_tool}` is not a valid tool."
                        f"Pick one from [{', '.join(tool_names)}]."
                    )
                )

                continue

            # Tool is valid. Run it
            tools_used.append(selected_tool)

            # Get tool input
            tool_fn = next(tool for tool in tools if tool.__name__ == selected_tool)
            tool_input = parsed_tools.get("tool_input") or {}

            # Run tool
            tool_result = await tool_fn(**tool_input)

            # Yield the observation
            yield User(content=f"Tool Result: {tool_result.strip()}")

        # Yield final response if any
        response = parsed_tools.get(
            "final_response",
            "I am not sure how to answer this question",
        )

        yield Result(content=dict(response=response, tools_used=tools_used))



from pprint import pformat
from typing import Any


def pprint_color(obj: Any) -> Any:
    """Pretty-print in color."""

    from pygments import highlight
    from pygments.formatters import Terminal256Formatter
    from pygments.lexers import PythonLexer

    print(highlight(pformat(obj), PythonLexer(), Terminal256Formatter()), end="")
    return obj
